Package: scanCP
Type: Package
Title: Deep Learning–Based Changepoint Detection with Local Neural
        Models
Version: 0.1.0
Authors@R: c(
    person(
      "Arman", "Azizyan",
      email = "arman.azizyan@gmail.com",
      role = c("aut", "cre")),
    person(
      "Abolfazl", "Safikhani",
      role = "aut"))
Description: Implementation of deep learning–based changepoint detection 
    algorithm designed for time series with smooth local fluctuations. 
    The method fits localized feed‑forward neural networks to approximate the 
    underlying smooth component and constructs a residual‑based detector that 
    isolates abrupt structural changes. A fully data‑adaptive 
    empirical cumulative distribution function (ECDF) based thresholding 
    rule and refinement procedures yield accurate changepoint localization 
    without parametric assumptions on noise or trend structure.
License: GPL-2
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: plotly, RSNNS, foreach, doSNOW, parallel, pracma, stats,
        magrittr, tidyr
NeedsCompilation: no
Packaged: 2026-05-27 18:21:15 UTC; arman
Author: Arman Azizyan [aut, cre],
  Abolfazl Safikhani [aut]
Maintainer: Arman Azizyan <arman.azizyan@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-30 13:40:08 UTC
Built: R 4.6.0; ; 2026-05-30 17:14:02 UTC; unix

# Generated by roxygen2: do not edit by hand

export(best_split_free)
export(calc_detector)
export(combine_detectors)
export(decompose_signal_core)
export(detect_cp_ecdf)
export(fit_global_mlp)
export(fit_mlp)
export(ma)
export(plot_mlp_fits_interactive)
export(scan_cp)
export(scan_cp_multi_async)
export(scan_cp_multi_sync)
export(select_best_spike)
export(simulate_piecewise_signal_idx)
import(RSNNS)
import(doSNOW)
import(foreach)
import(parallel)
importFrom(RSNNS,mlp)
importFrom(magrittr,"%>%")
importFrom(plotly,add_lines)
importFrom(plotly,layout)
importFrom(plotly,plot_ly)
importFrom(pracma,findpeaks)
importFrom(stats,ecdf)
importFrom(stats,filter)
importFrom(stats,kmeans)
importFrom(stats,median)
importFrom(stats,rnorm)
importFrom(tidyr,replace_na)
importFrom(utils,modifyList)
importFrom(utils,setTxtProgressBar)
importFrom(utils,txtProgressBar)

# scanCP: Multilayer Perceptron–Based Changepoint Detection

`scanCP` provides a fast, robust, and fully automated changepoint detection
pipeline based on rolling multilayer perceptron (MLP) models. The method
combines local and global structure learning, RSS-based detector statistics,
ECDF thresholding, and local refinement to identify structural breaks in
univariate signals.

The package is designed for signals with abrupt level shifts, smooth trends,
and heterogeneous noise. It includes tools for detector construction,
changepoint extraction, signal correction, and visualization.

---

## Features

- Rolling-window MLP fits (window sizes `w` and `2w`)
- RSS-based hybrid detector (ratio + difference)
- Automatic threshold selection via spacing-curve analysis
- Local refinement using k-means segmentation
- Piecewise-constant correction of the signal
- Optional global MLP smoother
- Parallel computation for speed
- Clean, modular API

---

## Installation

### From CRAN (when available)

```r
install.packages("scanCP")
```

### From GitHub (development version)

```r
# install.packages("devtools")
devtools::install_github("armanazizyan/scanCP")
```

---

## Quick Start

```r
library(scanCP)

# Simulate a simple piecewise-constant signal
set.seed(123)
y <- c(
  rnorm(200, 0, 1),
  rnorm(200, 3, 1),
  rnorm(200, -2, 1),
  rnorm(200, 1, 1)
)

# Run the full changepoint pipeline
res <- scan_cp(y, w = 100)

# Extract changepoints
res$changepoints

# Plot the corrected signal
plot(y, type = "l", col = "gray70")
lines(res$corrected_signal, col = "blue", lwd = 2)
abline(v = res$changepoints, col = "red", lwd = 2)
```

---

## Method Overview

`scanCP` detects changepoints using a hybrid detector statistic derived from
rolling MLP fits:

- A small-window MLP (size `w`) models local structure.
- A large-window MLP (size `2w`) models broader structure.
- For each position, the detector combines:
  - a ratio of RSS values
  - a difference of RSS values

The detector is smoothed, thresholded using the ECDF or spacing-curve
analysis, and refined using a local k-means split. The result is a set of
precise changepoint locations and a corrected signal.

---

## Customizing the MLP

All MLP hyperparameters can be controlled via `mlp_control`:

```r
res <- scan_cp(
  y,
  w = 100,
  mlp_control = list(
    hl1 = 10,
    hl2 = 20,
    ep1 = 1500,
    act1 = "Act_Tanh"
  )
)
```

---

## Documentation

Full function documentation is available via:

```r
help(package = "scanCP")
```

---

## License

GPL-2 License © Arman Azizyan

---

## Contributing

Issues and pull requests are welcome at:

https://github.com/armanazizyan/scanCP

#' Asynchronous Multivariate Changepoint Detection
#'
#' @description
#' Applies the univariate \code{\link{scan_cp}} pipeline independently to
#' each column of a multivariate signal. Each dimension is processed
#' separately (asynchronously), producing a list of univariate results.
#'
#' @param Y Numeric matrix (n × p). Each column is a signal dimension.
#' @param w Integer. Window size for rolling MLPs and detector construction.
#'   Defaults to 100.
#' @param ma_window Integer. Moving-average smoothing window for the detector.
#'   Defaults to \code{w}.
#' @param threshold Either \code{"auto"} (default) or a numeric ECDF threshold.
#' @param threshold_tails Numeric vector of length 2 giving tail cutoff values
#' for automatic thresholding.
#'   Defaults to \code{c(0.6, 0.95)}.
#' @param min_cp_distance Integer. Minimum distance between detected
#'   changepoints. Defaults to \code{2 * w}.
#' @param margin Integer. Local refinement margin. Defaults to \code{floor(w/2)}.
#' @param use_abs_det Logical. Whether to use \code{abs(detector)} before
#'   smoothing. Defaults to \code{TRUE}.
#' @param mlp_control List of parameters passed to \code{\link{fit_mlp}}.
#'
#' @return A named list of length \code{p}, where each element is the output
#'   of \code{\link{scan_cp}} applied to the corresponding column of \code{Y}.
#'
#' @examples
#' \donttest{
#'   Y <- cbind(
#'     x1 = c(rnorm(200), rnorm(200, 3)),
#'     x2 = c(rnorm(200), rnorm(200, -2))
#'   )
#'
#'   res <- scan_cp_multi_async(Y, w = 100)
#'   res$x1$changepoints
#'   res$x2$changepoints
#' }
#'
#' @export
scan_cp_multi_async <- function(
    Y,
    w = 100,
    ma_window = w,
    threshold = "auto",
    threshold_tails = c(0.6, 0.95),
    min_cp_distance = 2 * w,
    margin = floor(w / 2),
    use_abs_det = TRUE,
    mlp_control = list()
) {

  if (!is.matrix(Y))
    stop("Y must be a numeric matrix.")

  p <- ncol(Y)
  out <- vector("list", p)
  names(out) <- colnames(Y) %||% paste0("dim", seq_len(p))

  for (j in seq_len(p)) {
    out[[j]] <- scan_cp(
      y = Y[, j],
      w = w,
      ma_window = ma_window,
      threshold = threshold,
      threshold_tails = threshold_tails,
      min_cp_distance = min_cp_distance,
      margin = margin,
      use_abs_det = use_abs_det,
      mlp_control = mlp_control
    )
  }

  out
}

#' Fit Rolling MLP Models for Change Point Detection
#'
#' @description
#' Fits two multilayer perceptron (MLP) models over rolling windows of a
#' univariate signal. A small-window MLP (window size \code{w}) captures
#' local structure, while a large-window MLP (window size \code{2w})
#' captures broader trends. Their residual behavior is used by
#' \code{\link{calc_detector}} to construct the changepoint detector.
#'
#' All MLP hyperparameters are supplied through a unified
#' \code{mlp_control} list for consistency with \code{\link{scan_cp}}.
#'
#' @param vec Numeric vector. The input signal.
#' @param w Integer. Window size for the small-window MLP. The large-window
#'   MLP automatically uses window size \code{2w}.
#' @param mlp_control A named list of MLP hyperparameters. Any subset may be
#'   supplied; unspecified values fall back to defaults. Supported fields:
#'   \describe{
#'     \item{\code{act1}}{Activation function for the small-window MLP
#'       (\code{"Act_Logistic"}).}
#'     \item{\code{act2}}{Activation function for the large-window MLP
#'       (\code{"Act_Logistic"}).}
#'     \item{\code{lr1}}{Learning rate for the small-window MLP
#'       (default: \code{0.001}).}
#'     \item{\code{lr2}}{Learning learning rate for the large-window MLP
#'       (default: \code{0.001}).}
#'     \item{\code{hl1}}{Number of hidden units for the small-window MLP
#'       (default: \code{6}).}
#'     \item{\code{hl2}}{Number of hidden units for the large-window MLP
#'       (default: \code{8}).}
#'     \item{\code{ep1}}{Training epochs for the small-window MLP
#'       (default: \code{300}).}
#'     \item{\code{ep2}}{Training epochs for the large-window MLP
#'       (default: \code{600}).}
#'   }
#'
#' @param parallel Logical. If \code{TRUE}, rolling-window MLPs are
#'   fitted using parallel computation via \pkg{parallel}, \pkg{foreach},
#'   and \pkg{doSNOW}. If \code{FALSE}, all computation is performed
#'   serially. Setting \code{parallel = FALSE} is recommended for CRAN checks
#'   and for systems where parallel backends are unavailable.
#'
#' @return A list with two elements:
#' \describe{
#'   \item{\code{small}}{List of fitted values for each rolling window of size \code{w}.}
#'   \item{\code{large}}{List of fitted values for each rolling window of size \code{2w}.}
#' }
#'
#' @details
#' Each rolling window is standardized before fitting. The returned matrices
#' contain three columns:
#' \enumerate{
#'   \item standardized input \code{x}
#'   \item fitted values \code{y_hat}
#'   \item original indices
#' }
#'
#' When \code{parallel = TRUE}, a cluster is created using
#' \code{parallel::makeCluster()}.
#'
#' When \code{parallel = FALSE}, the function falls back to
#' serial execution using \code{\%do\%}.
#'
#' @import RSNNS
#' @import foreach
#' @import doSNOW
#' @import parallel
#' @importFrom utils modifyList txtProgressBar setTxtProgressBar
#' @export
fit_mlp <- function(
    vec,
    w = 100,
    mlp_control = list(),
    parallel = FALSE
) {

  # Defaults
  default_mlp_control <- list(
    act1 = "Act_Logistic",
    act2 = "Act_Logistic",
    lr1  = 0.001,
    lr2  = 0.001,
    hl1  = 6,
    hl2  = 8,
    ep1  = 300,
    ep2  = 600
  )

  mlp <- modifyList(default_mlp_control, mlp_control)

  # Extract
  act1 <- mlp$act1; act2 <- mlp$act2
  lr1  <- mlp$lr1;  lr2  <- mlp$lr2
  hl1  <- mlp$hl1;  hl2  <- mlp$hl2
  ep1  <- mlp$ep1;  ep2  <- mlp$ep2

  n.val <- length(vec)
  x <- 1:n.val
  y <- vec

  # -----------------------------
  # Parallel or serial backend
  # -----------------------------
  if (parallel) {
    num_cores <- max(1, parallel::detectCores() - 1)
    cl <- parallel::makeCluster(num_cores)
    doSNOW::registerDoSNOW(cl)

    pb <- txtProgressBar(min = 0, max = n.val - w, style = 3)
    progress <- function(n) setTxtProgressBar(pb, n)
    opts <- list(progress = progress)

    `%dopar_or_do%` <- `%dopar%`
  } else {
    # Serial mode: no cluster, no progress bar
    opts <- list()
    `%dopar_or_do%` <- `%do%`
  }

  # -----------------------------
  # Small-window MLP
  # -----------------------------
  if (parallel) message("Fitting small-window MLPs...")

  res.small <- foreach::foreach(
    idx = 1:(n.val - w + 1),
    .packages = "RSNNS",
    .options.snow = opts
  ) %dopar_or_do% {

    i <- idx
    sub_x <- as.matrix(scale(x[i:(i + w - 1)]))
    sub_y <- as.matrix(y[i:(i + w - 1)])

    net <- RSNNS::mlp(
      sub_x, sub_y,
      size = hl1,
      learnFuncParams = lr1,
      maxit = ep1,
      linOut = TRUE,
      hiddenActFunc = act1
    )

    cbind(sub_x, net$fitted.values, i:(i + w - 1))
  }

  if (parallel) close(pb)

  # -----------------------------
  # Large-window MLP
  # -----------------------------
  if (parallel) {
    pb <- txtProgressBar(min = 0, max = n.val - 2 * w, style = 3)
    progress <- function(n) setTxtProgressBar(pb, n)
    opts <- list(progress = progress)
  }

  if (parallel) message("Fitting large-window MLPs...")

  res.large <- foreach::foreach(
    idx = 1:(n.val - 2 * w + 1),
    .packages = "RSNNS",
    .options.snow = opts
  ) %dopar_or_do% {

    i <- idx
    sub_x <- as.matrix(scale(x[i:(i + 2 * w - 1)]))
    sub_y <- as.matrix(y[i:(i + 2 * w - 1)])

    net <- RSNNS::mlp(
      sub_x, sub_y,
      size = hl2,
      learnFuncParams = lr2,
      maxit = ep2,
      linOut = TRUE,
      hiddenActFunc = act2
    )

    cbind(sub_x, net$fitted.values, i:(i + 2 * w - 1))
  }

  if (parallel) {
    close(pb)
    parallel::stopCluster(cl)
  }

  list(
    small = res.small,
    large = res.large
  )
}

#' Core Signal Decomposition (Changepoints + Piecewise Correction)
#'
#' @description
#' Extracts changepoints from a detector statistic using ECDF thresholding,
#' spacing-curve analysis, and local refinement. Produces the corrected
#' signal and its piecewise-constant component. This function does *not*
#' perform global MLP smoothing.
#'
#' @param y Numeric vector. Original signal.
#' @param detector Numeric vector. Detector statistic produced by
#'   \code{\link{calc_detector}}.
#' @param w Integer. Window size used in detector construction.
#' @param ma_window Integer. Moving-average smoothing window for the detector
#'   and spacing curve. Defaults to \code{w}.
#' @param right_tail_cutoff Numeric. Upper ECDF cutoff for automatic
#'   threshold selection. Defaults to \code{0.95}.
#' @param left_tail_cutoff Numeric. Lower ECDF cutoff for automatic
#'   threshold selection. Defaults to \code{0.6}.
#' @param threshold Either \code{"auto"} (default) or a numeric ECDF
#'   threshold for selecting significant peaks.
#' @param use_abs_det Logical. Whether to use \code{abs(detector)} before
#'   peak detection. Defaults to \code{TRUE}.
#' @param min_cp_distance Integer. Minimum distance between detected peaks.
#'   Defaults to \code{2 * w}.
#' @param margin Integer. Refinement margin around each changepoint.
#'   Default \code{floor(w / 2)}.
#' @param circular Logical. Whether moving-average smoothing wraps around.
#'
#' @return A list containing:
#' \describe{
#'   \item{corrected_signal}{The refined, piecewise-corrected signal.}
#'   \item{piecewise_constant}{Estimated piecewise-constant component.}
#'   \item{raw_correction}{Cumulative correction vector.}
#'   \item{changepoints}{Refined changepoint locations.}
#'   \item{changepoint_ecdf}{ECDF values of selected peaks.}
#'   \item{shift_values}{Estimated shifts at each changepoint.}
#'   \item{smoothed_detector}{Smoothed detector statistic.}
#'   \item{ecdf_spacing}{Spacing-curve values (if \code{threshold = "auto"}).}
#'   \item{threshold_used}{Final threshold applied.}
#'   \item{local_maxima}{Matrix of detected local maxima.}
#' }
#'
#' @details
#' The function identifies local maxima of the smoothed detector, evaluates
#' their significance using the ECDF of the detector, optionally computes a
#' spacing curve for automatic threshold selection, and refines each
#' changepoint using a two-cluster k-means split within a local window.
#'
#' @importFrom pracma findpeaks
#' @importFrom stats ecdf median kmeans rnorm
#' @export
decompose_signal_core <- function(
    y,
    detector,
    w = 200,
    ma_window = 100,
    right_tail_cutoff = 0.95,
    left_tail_cutoff  = 0.6,
    threshold = "auto",
    use_abs_det = TRUE,
    min_cp_distance = NULL,
    circular = FALSE,
    margin = NULL
) {

  # --- 0. Preprocessing -------------------------------------------------------

  if (use_abs_det)
    detector <- abs(detector)

  n <- length(y)

  if (is.null(min_cp_distance))
    min_cp_distance <- 2 * w

  if (is.null(margin))
    margin <- floor(w / 2)

  # --- 1. Local maxima of raw detector ---------------------------------------

  # (Removed raw_peaks — unused)

  # --- 2. Smooth detector -----------------------------------------------------

  sm.det <- as.numeric(ma(detector, n = ma_window, circular = circular))
  sm.det[is.na(sm.det)] <- 0

  # --- 3. Local maxima of smoothed detector ----------------------------------

  l.max <- pracma::findpeaks(sm.det, minpeakdistance = min_cp_distance)
  if (is.null(l.max) || nrow(l.max) == 0) {
    return(list(
      corrected_signal   = y,
      piecewise_constant = rep(0, n),
      raw_correction     = rep(0, n),
      changepoints       = integer(0),
      changepoint_ecdf   = numeric(0),
      shift_values       = numeric(0),
      smoothed_detector  = sm.det,
      ecdf_spacing       = NULL,
      threshold_used     = NA,
      local_maxima       = NULL
    ))
  }

  # --- 4. ECDF significance ---------------------------------------------------

  f_det <- stats::ecdf(sm.det)
  ecdf_vals <- f_det(l.max[, 1])
  l.max <- cbind(l.max, ecdf_vals)  # col 5 = ECDF

  # --- 5. Spacing curve (only if needed) -------------------------------------

  if (is.character(threshold) && threshold == "auto") {
    x.det <- sort(sm.det)
    dx <- diff(x.det)
    s <- as.numeric(ma(dx, n = ma_window, circular = circular))
    s[is.na(s)] <- 0

    thr <- select_best_spike(
      s,
      right_tail_cutoff = right_tail_cutoff,
      left_tail_cutoff  = left_tail_cutoff
    )
  } else if (is.numeric(threshold)) {
    thr <- threshold
    s <- NULL
  } else {
    stop("threshold must be 'auto' or numeric.")
  }

  # --- 6. Changepoints (raw) --------------------------------------------------

  cps_raw <- l.max[l.max[, 5] >= thr, , drop = FALSE]

  if (nrow(cps_raw) == 0) {
    return(list(
      corrected_signal   = y,
      piecewise_constant = rep(0, n),
      raw_correction     = rep(0, n),
      changepoints       = integer(0),
      changepoint_ecdf   = numeric(0),
      shift_values       = numeric(0),
      smoothed_detector  = sm.det,
      ecdf_spacing       = s,
      threshold_used     = thr,
      local_maxima       = l.max
    ))
  }

  cps <- cps_raw[, 2] + w

  # --- 7. Local refinement ----------------------------------------------------

  refined <- c()
  shifts <- c()

  for (cp in sort(cps)) {

    lo <- max(1, cp - margin)
    hi <- min(n, cp + margin)

    segment <- y[lo:hi]

    b <- kmeans(segment, centers = 2)
    bsf <- best_split_free(b$cluster)

    corrected_cp <- cp + (margin - bsf$index)
    corrected_cp <- min(max(corrected_cp, 1), n)
    refined <- c(refined, corrected_cp)

    left_idx  <- max(1, corrected_cp - margin) : corrected_cp
    right_idx <- corrected_cp : min(n, corrected_cp + margin)

    shifts <- c(
      shifts,
      stats::median(y[left_idx]) - stats::median(y[right_idx])
    )
  }

  # --- 8. Build correction vector --------------------------------------------

  cps <- refined
  starts <- cps + 1
  ends <- c(cps[-1], n)

  cor.vec <- numeric(n)
  cum.shifts <- cumsum(shifts)

  for (i in seq_along(starts)) {
    cor.vec[starts[i]:ends[i]] <- cum.shifts[i]
  }

  new.y <- y + cor.vec

  # --- 9. Piecewise constant component ---------------------------------------

  idx_start <- c(1, cps + 1)
  idx_end   <- c(cps, n)

  step_mean <- numeric(n)
  for (i in seq_along(idx_start)) {
    step_mean[idx_start[i]:idx_end[i]] <-
      mean((y - new.y)[idx_start[i]:idx_end[i]])
  }

  # --- 10. Return -------------------------------------------------------------

  list(
    corrected_signal   = new.y,
    piecewise_constant = step_mean,
    raw_correction     = cor.vec,
    changepoints       = cps,
    changepoint_ecdf   = cps_raw[, 5],
    shift_values       = shifts,
    smoothed_detector  = sm.det,
    ecdf_spacing       = s,
    threshold_used     = thr,
    local_maxima       = l.max
  )
}




#' Fit Global MLP and Compute Residual Diagnostics
#'
#' @description
#' Fits a global MLP model to a corrected signal and computes:
#' - smooth trend estimate
#' - residuals
#'
#' This function does NOT compute a piecewise-constant component.
#' That is handled entirely in `decompose_signal_core()`.
#'
#' @param corrected_signal Numeric vector. Output of [decompose_signal_core()].
#'
#' @return A list containing:
#' \describe{
#'   \item{smooth_curve}{Global MLP smooth fit.}
#'   \item{residual}{Residuals from the smooth fit.}
#'   \item{mlp_model}{The fitted RSNNS MLP model.}
#' }
#'
#' @importFrom RSNNS mlp
#' @export
fit_global_mlp <- function(corrected_signal) {

  n <- length(corrected_signal)
  x <- 1:n

  # Fit global MLP
  sub_x <- as.matrix(scale(x))
  sub_y <- as.matrix(corrected_signal)

  net <- RSNNS::mlp(
    sub_x, sub_y,
    size = 64,
    learnFuncParams = c(0.01, 0.001),
    maxit = 2000,
    linOut = TRUE,
    hiddenActFunc = "Act_Logistic"
  )

  smooth <- net$fitted.values
  resid <- corrected_signal - smooth

  list(
    smooth_curve = smooth,
    residual     = resid,
    mlp_model    = net
  )
}

#' Best Split for Binary Labels (Free Assignment)
#'
#' @description
#' Given a sequence of labels in \{1, 2\}, this function finds the best split
#' point that maximizes the number of correctly assigned labels on each side.
#' The left and right segments are each assigned their majority class.
#'
#' @param labels Integer vector containing only 1 and 2.
#'
#' @return A list with:
#' \describe{
#'   \item{index}{The optimal split location (between index and index + 1).}
#'   \item{total_correct}{Maximum number of correctly assigned labels.}
#'   \item{left_majority}{Majority class on the left segment.}
#'   \item{right_majority}{Majority class on the right segment.}
#'   \item{accuracy_if_assigning_majorities}{Proportion correctly assigned.}
#' }
#'
#' @details
#' For each possible split point \(i\), the function computes:
#' \itemize{
#'   \item left correct = max(#1_left, #2_left)
#'   \item right correct = max(#1_right, #2_right)
#' }
#' and selects the split maximizing their sum.
#'
#' @examples
#' labels <- c(1,1,1,2,2,2)
#' best_split_free(labels)
#'
#' @export
best_split_free <- function(labels) {

  # Validate
  if (length(labels) < 2)
    stop("Need at least 2 labels to split.")

  if (any(is.na(labels)))
    stop("Labels contain NA; please remove or impute.")

  u <- sort(unique(labels))
  if (!all(u %in% c(1, 2)))
    stop("Labels must be in {1, 2}.")

  n <- length(labels)

  # Cumulative counts
  cum1 <- cumsum(labels == 1)
  cum2 <- cumsum(labels == 2)

  total1 <- cum1[n]
  total2 <- cum2[n]

  # Left and right correct counts
  left_correct  <- pmax(cum1, cum2)
  right_correct <- pmax(total1 - cum1, total2 - cum2)

  # Only splits 1..(n-1)
  total_correct <- left_correct[1:(n - 1)] + right_correct[1:(n - 1)]

  # Best split
  best_i <- which.max(total_correct)

  list(
    index = best_i,
    total_correct = total_correct[best_i],
    left_majority  = if (cum1[best_i] >= cum2[best_i]) 1 else 2,
    right_majority = if ((total1 - cum1[best_i]) >= (total2 - cum2[best_i])) 1 else 2,
    accuracy_if_assigning_majorities = total_correct[best_i] / n
  )
}


#' Select the Most Significant Spike in an ECDF Spacing Curve
#'
#' @description
#' Identifies the most prominent spike in a spacing curve derived from the
#' ECDF of a smoothed detector statistic. This is used to automatically
#' determine a significance threshold for changepoint selection.
#'
#' @param s Numeric vector. Smoothed spacing curve.
#' @param right_tail_cutoff Numeric in (0,1). Exclude spikes with ECDF
#'   probability above this value (typically near 1).
#' @param left_tail_cutoff Numeric in (0,1). Exclude spikes with ECDF
#'   probability below this value (to avoid trivial early spikes).
#'
#' @return A numeric value in (0,1) representing the selected ECDF threshold,
#'   or `NA` if no valid spike is found.
#'
#' @details
#' Uses `pracma::findpeaks()` to identify local maxima. Prominence is computed
#' as:
#' \deqn{ peak_{height} - \max(left_{base}, right_{base}) }{}
#'
#' The spike with the largest prominence within the allowed ECDF range is
#' selected.
#'
#' @importFrom pracma findpeaks
#'
#' @examples
#' s <- runif(100)
#' select_best_spike(s)
#'
#' @usage select_best_spike(s,
#'                 right_tail_cutoff = 0.95,
#'                 left_tail_cutoff  = 0.6)
#'
#' @export
select_best_spike <- function(
    s,
    right_tail_cutoff = 0.95,
    left_tail_cutoff  = 0.6
) {

  n <- length(s)
  p <- (1:n) / (n + 1)   # ECDF probabilities

  # Find peaks
  peaks <- pracma::findpeaks(s, sortstr = FALSE)
  if (is.null(peaks))
    return(NA)

  peak_heights <- peaks[, 1]
  peak_idx     <- peaks[, 2]
  left_base    <- peaks[, 3]
  right_base   <- peaks[, 4]

  # Prominence
  baseline <- pmax(s[left_base], s[right_base])
  prominence <- peak_heights - baseline

  # Exclude tail regions
  valid <- (p[peak_idx] < right_tail_cutoff) &
    (p[peak_idx] > left_tail_cutoff)

  if (!any(valid))
    return(NA)

  # Select spike with largest prominence
  best_idx <- peak_idx[valid][which.max(prominence[valid])]
  best_p   <- p[best_idx]

  best_p
}

#' Interactive Plot of Rolling MLP Fits
#'
#' @description
#' Creates an interactive Plotly visualization showing multiple rolling-window
#' MLP fits over a univariate signal. A slider allows the user to move through
#' different window positions and compare model fits dynamically.
#'
#' @param y Numeric vector. The original signal.
#' @param fit_mlp_res A list containing rolling MLP fit results. Expected to be
#'   a list of two lists, each containing matrices with fitted values.
#' @param t.chp.ind Optional numeric vector of changepoint indices to draw as
#'   vertical dashed lines.
#' @param step Integer. Step size for the slider increments.
#' @param start Integer. Starting index for the first model window.
#' @param w Integer. Window size used in the rolling MLP fitting.
#'
#' @return A Plotly object.
#'
#' @importFrom plotly plot_ly add_lines layout
#' @importFrom magrittr %>%
#' @export
plot_mlp_fits_interactive <- function(
    y,
    fit_mlp_res,
    t.chp.ind = NA,
    step = 50,
    start = 75,
    w = 100
) {

  # ------------------------------------------------------------
  # Basic setup
  # ------------------------------------------------------------
  n <- length(y)

  # fit_mlp_res structure:
  #   fit_mlp_res[[1]] = model 2 + model 3
  #   fit_mlp_res[[2]] = model 1
  res.list1 <- fit_mlp_res[[2]]               # Model 1 fits
  res.list2 <- fit_mlp_res[[1]]               # Model 2 fits
  res.list3 <- fit_mlp_res[[1]][w + 1:n]      # Model 3 fits (shifted)

  # Indices where models were fitted
  idx_list <- seq(start, n - 2 * w, by = step)
  n_models <- length(idx_list)

  # ------------------------------------------------------------
  # Changepoint vertical lines (optional)
  # ------------------------------------------------------------
  if (!is.na(t.chp.ind)) {
    cp_segments <- lapply(t.chp.ind, function(cp) {
      list(
        type = "line",
        x0 = cp, x1 = cp,
        y0 = min(y), y1 = max(y),
        line = list(color = "blue", dash = "dash")
      )
    })
  } else {
    cp_segments <- NULL
  }

  # ------------------------------------------------------------
  # Initial plot (frame 1)
  # ------------------------------------------------------------
  i0 <- idx_list[1]

  p <- plotly::plot_ly() %>%
    # Original signal
    plotly::add_lines(
      x = 1:n, y = y,
      name = "Original Signal",
      line = list(color = "rgba(0,0,0,0.3)")
    ) %>%

    # Model 1
    plotly::add_lines(
      x = res.list1[[i0]][, 3],
      y = res.list1[[i0]][, 2],
      name = "Model 1",
      line = list(color = "red", width = 3)
    ) %>%

    # Model 2
    plotly::add_lines(
      x = res.list2[[i0]][, 3],
      y = res.list2[[i0]][, 2],
      name = "Model 2",
      line = list(color = "green", width = 3)
    ) %>%

    # Model 3
    plotly::add_lines(
      x = res.list3[[i0]][, 3],
      y = res.list3[[i0]][, 2],
      name = "Model 3",
      line = list(color = "purple", width = 3)
    ) %>%

    # Layout with slider placeholder
    plotly::layout(
      title = "Multiple MLP Fits - Interactive Slider",
      xaxis = list(title = "Index"),
      yaxis = list(title = "Value"),
      shapes = cp_segments,
      sliders = list(list(
        active = 0,
        currentvalue = list(prefix = "Model index: "),
        steps = lapply(seq_len(n_models), function(k) {
          list(
            label = as.character(idx_list[k]),
            method = "animate",
            args = list(
              list(paste0("frame", k)),
              list(
                mode = "immediate",
                frame = list(duration = 0, redraw = FALSE),
                transition = list(duration = 0)
              )
            )
          )
        })
      ))
    )

  # ------------------------------------------------------------
  # Animation frames
  # ------------------------------------------------------------
  frames <- lapply(seq_len(n_models), function(k) {
    i <- idx_list[k]

    list(
      name = paste0("frame", k),
      data = list(
        NULL,  # placeholder for original signal (unchanged)
        list(  # Model 1
          x = res.list1[[i]][, 3],
          y = res.list1[[i]][, 2],
          mode = "lines",
          line = list(color = "red", width = 3)
        ),
        list(  # Model 2
          x = res.list2[[i]][, 3],
          y = res.list2[[i]][, 2],
          mode = "lines",
          line = list(color = "green", width = 3)
        ),
        list(  # Model 3
          x = res.list3[[i]][, 3],
          y = res.list3[[i]][, 2],
          mode = "lines",
          line = list(color = "purple", width = 3)
        )
      )
    )
  })

  # Attach frames to the plotly object
  p$x$frames <- frames

  p
}

#' Compute the MLP-Based Changepoint Detector Statistic
#'
#' @description
#' Computes the hybrid detector statistic used in scanCP. The detector
#' combines a *ratio* component and a *difference* component derived from
#' the residual sum of squares (RSS) of:
#'
#' - two adjacent small-window MLP fits (window size `w`)
#' - one large-window MLP fit (window size `2w`)
#'
#' The detector highlights locations where the large-window fit performs
#' substantially worse than the two small-window fits, indicating a
#' potential changepoint.
#'
#' @param y Numeric vector. Original signal.
#' @param fit_mlp_res List. Output of \code{\link{fit_mlp}}, containing:
#'   \itemize{
#'     \item \code{small} — list of small-window MLP predictions
#'     \item \code{large} — list of large-window MLP predictions
#'   }
#' @param w Integer. Window size used for the small-window MLP.
#' @param ma_window Integer. Moving-average smoothing window applied to the
#'   detector. Defaults to \code{w}.
#' @param use_abs_det Logical. Whether to take \code{abs(detector)} before
#'   smoothing. Defaults to \code{TRUE}.
#'
#' @return A numeric vector of detector values of length \code{n - 2w}.
#'
#' @details
#' For each index \code{i}, the detector uses:
#'
#' - \code{rss1}: RSS of the small-window MLP on \code{[i, i+w-1]}
#' - \code{rss2}: RSS of the small-window MLP on \code{[i+w, i+2w-1]}
#' - \code{rss.tot}: RSS of the large-window MLP on \code{[i, i+2w-1]}
#'
#' The ratio component is:
#' \deqn{ (rss.tot + b) / (rss1 + rss2 + b) }
#'
#' The difference component is:
#' \deqn{ rss.tot - rss1 - rss2 }
#'
#' The final detector is a convex combination:
#' \deqn{ (1 - a) \cdot \text{ratio} + a \cdot \text{difference} }
#'
#' Internal parameters \code{a}, \code{b}, and \code{scale_01} are kept
#' fixed at their defaults for simplicity and stability.
#'
#' @examples
#' \donttest{
#'   y <- rnorm(300)
#'   fit <- fit_mlp(y, w = 100)
#'   det <- calc_detector(y, fit, w = 100)
#' }
#'
#' @usage calc_detector(y, fit_mlp_res, w = 100,
#'               ma_window = w, use_abs_det = TRUE)
#'
#' @export
calc_detector <- function(
    y,
    fit_mlp_res,
    w = 100,
    ma_window = w,
    use_abs_det = TRUE
) {

  scale_01 = FALSE
  circular = FALSE
  a = 1
  b = 0

  # Validate inputs
  if (!is.numeric(y)) stop("y must be numeric.")
  if (!is.list(fit_mlp_res)) stop("fit_mlp_res must be a list.")
  if (!all(c("small", "large") %in% names(fit_mlp_res)))
    stop("fit_mlp_res must contain elements 'small' and 'large'.")

  res.small <- fit_mlp_res$small
  res.large <- fit_mlp_res$large

  n.val <- length(y)
  n.det <- n.val - 2 * w
  if (n.det <= 0) stop("Signal too short for window size w.")

  diff1 <- numeric(n.det)   # ratio component
  diff2 <- numeric(n.det)   # difference component

  for (i in seq_len(n.det)) {

    rss1 <- sum((y[i:(i + w - 1)] - res.small[[i]][, 2])^2)
    rss2 <- sum((y[(i + w):(i + 2 * w - 1)] - res.small[[i + w]][, 2])^2)
    rss.tot <- sum((y[i:(i + 2 * w - 1)] - res.large[[i]][, 2])^2)

    diff1[i] <- (rss.tot + b) / (rss1 + rss2 + b)
    diff2[i] <- rss.tot - rss1 - rss2
  }

  # Combine components
  det <- (1 - a) * diff1 + a * diff2

  # Optional scaling
  if (scale_01) {
    det <- (det - min(det)) / (max(det) - min(det))
  }

  # Optional absolute value
  if (use_abs_det)
    det <- abs(det)

  # Smooth detector
  det <- as.numeric(ma(det, n = ma_window, circular = circular))
  det[is.na(det)] <- 0

  det
}

#' Combine Multiple Detector Statistics into a Joint Detector
#'
#' @description
#' Combines several univariate detector vectors into a single joint detector
#' using L1, L2, or max aggregation. Optionally applies moving-average
#' smoothing and computes contribution weights.
#'
#' @param det.lst List of numeric detector vectors (all same length).
#' @param method One of `"L1"`, `"L2"`, `"max"`.
#' @param ma_window Optional integer. If provided, each detector is smoothed
#'   using `ma()` with this window size.
#' @param circular Logical. Whether smoothing wraps around.
#' @param scale_contributions Logical. Whether to compute contribution weights.
#' @param scale_joint Logical. Whether to scale the joint detector to \[0,1\].
#'
#' @return A list containing:
#' \describe{
#'   \item{joint}{Joint detector vector.}
#'   \item{s.mat}{Scaled detector matrix.}
#'   \item{det.mat}{Raw detector matrix.}
#'   \item{contrib}{Contribution weights (or NULL).}
#' }
#'
#' @importFrom tidyr replace_na
#' @export
combine_detectors <- function(
    det.lst,
    method = c("L1", "L2", "max"),
    ma_window = NULL,
    circular = FALSE,
    scale_contributions = TRUE,
    scale_joint = TRUE
) {

  method <- match.arg(method)

  # 1. Optional smoothing
  if (!is.null(ma_window)) {
    det.lst <- lapply(det.lst, ma, n = ma_window, circular = circular)
  }

  # 2. Convert to matrix
  det.mat <- do.call(cbind, lapply(det.lst, as.vector))

  # 3. Scaling (currently identity)
  scale.mat <- det.mat

  # 4. Joint detector
  joint.det <- switch(
    method,
    L1  = rowSums(abs(scale.mat)),
    L2  = sqrt(rowSums(scale.mat^2)),
    max = apply(scale.mat, 1, max)
  )

  # 5. Scale joint detector
  if (scale_joint) {
    rng <- range(joint.det, na.rm = TRUE)
    joint.det <- (joint.det - rng[1]) / (rng[2] - rng[1])
  }

  # 6. Contribution weights
  if (scale_contributions) {
    contrib <- switch(
      method,
      L1  = abs(scale.mat),
      L2  = scale.mat^2,
      max = pmax(scale.mat, 0)
    )
    contrib <- contrib / rowSums(contrib)
  } else {
    contrib <- NULL
  }

  list(
    joint   = as.numeric(joint.det),
    s.mat   = scale.mat,
    det.mat = det.mat,
    contrib = contrib
  )
}

#' ECDF-Based Changepoint Detection for a Detector Statistic
#'
#' @description
#' Applies the ECDF + spacing-curve thresholding method to a detector vector
#' and returns raw changepoints (no correction).
#'
#' @param diff Numeric detector vector.
#' @param w Integer. Window size used in detector construction.
#' @param ma_window Integer. Moving-average window for smoothing.
#' @param right_tail_cutoff Numeric. ECDF upper cutoff.
#' @param left_tail_cutoff Numeric. ECDF lower cutoff.
#' @param threshold "auto" or numeric ECDF threshold.
#' @param circular Logical. Whether smoothing wraps around.
#'
#' @return A list containing:
#' \describe{
#'   \item{changepoints}{Raw changepoint indices.}
#'   \item{ecdf_values}{ECDF values of selected maxima.}
#'   \item{threshold_used}{Final threshold.}
#'   \item{local_maxima}{Matrix of local maxima.}
#'   \item{smoothed_detector}{Smoothed detector.}
#'   \item{ecdf_spacing}{Spacing curve.}
#' }
#'
#' @importFrom pracma findpeaks
#' @importFrom stats ecdf
#' @export
detect_cp_ecdf <- function(
    diff,
    w = 200,
    ma_window = 100,
    right_tail_cutoff = 0.95,
    left_tail_cutoff  = 0.6,
    threshold = "auto",
    circular = FALSE
) {

  # 1. Smooth detector
  ma.diff <- as.numeric(ma(diff, n = ma_window, circular = circular))

  # 2. Local maxima
  l.max <- pracma::findpeaks(replace_na(ma.diff, 0), minpeakdistance = 2 * w)

  if (is.null(l.max) || nrow(l.max) == 0) {
    return(list(
      changepoints      = integer(0),
      ecdf_values       = numeric(0),
      threshold_used    = NA,
      local_maxima      = NULL,
      smoothed_detector = ma.diff
    ))
  }

  # 3. ECDF significance
  f_diff <- stats::ecdf(ma.diff)
  ecdf_vals <- f_diff(l.max[, 1])
  l.max <- cbind(l.max, ecdf_vals)

  # 4. ECDF spacing curve
  x.diff <- sort(ma.diff)
  dx <- diff(x.diff)
  s <- as.numeric(ma(dx, n = w, circular = FALSE))
  s[is.na(s)] <- 0

  # 5. Threshold selection
  if (is.character(threshold) && threshold == "auto") {
    thr <- select_best_spike(
      s,
      right_tail_cutoff = right_tail_cutoff,
      left_tail_cutoff  = left_tail_cutoff
    )
  } else if (is.numeric(threshold)) {
    thr <- threshold
  } else {
    stop("threshold must be 'auto' or numeric.")
  }

  # 6. Raw changepoints
  cps_raw <- l.max[l.max[, 5] >= thr, , drop = FALSE]
  cps <- cps_raw[, 2] + w

  list(
    changepoints      = cps,
    ecdf_values       = cps_raw[, 5],
    threshold_used    = thr,
    local_maxima      = l.max,
    smoothed_detector = ma.diff,
    ecdf_spacing      = s
  )
}

#' Multivariate Synchronized Changepoint Detection
#'
#' @description
#' Implements the synchronized multivariate changepoint pipeline:
#' 1. Fit univariate MLP-based detectors for each dimension.
#' 2. Combine detectors into a joint detector using L1/L2/max.
#' 3. Detect changepoints on the joint detector.
#' 4. Compute per-dimension contributions at each detected CP.
#'
#' @param Y Numeric matrix (n × p). Each column is a signal dimension.
#' @param w Integer. Window size for rolling MLPs and detector.
#' @param method Character. Combination method: "L1", "L2", or "max".
#' @param mlp_params List of parameters passed to `fit_mlp()`.
#' @param detector_params List of parameters passed to `calc_detector()`.
#' @param combine_params List of parameters passed to `combine_detectors()`.
#' @param ecdf_params List of parameters passed to `detect_cp_ecdf()`.
#'
#' @return A list containing:
#' \describe{
#'   \item{detectors}{List of univariate detectors (one per dimension).}
#'   \item{joint_detector}{Joint detector vector.}
#'   \item{contributions}{Matrix of per-dimension contributions.}
#'   \item{changepoints}{Detected synchronized changepoints.}
#'   \item{cp_contributions}{Contribution of each dimension at each CP.}
#'   \item{ecdf_output}{Full output of `detect_cp_ecdf()`.}
#' }
#'
#' @export
scan_cp_multi_sync <- function(
    Y,
    w = 100,
    method = c("L1", "L2", "max"),
    mlp_params      = list(),
    detector_params = list(),
    combine_params  = list(),
    ecdf_params     = list()
) {

  method <- match.arg(method)

  if (!is.matrix(Y))
    stop("Y must be a numeric matrix.")

  p <- ncol(Y)
  n <- nrow(Y)

  # ------------------------------------------------------------
  # 1. Fit univariate detectors for each dimension
  # ------------------------------------------------------------
  det.lst <- vector("list", p)
  names(det.lst) <- colnames(Y) %||% paste0("dim", seq_len(p))

  for (j in seq_len(p)) {

    yj <- Y[, j]

    # Fit MLPs
    fit_args <- c(list(vec = yj, w = w), mlp_params)
    fit_res <- do.call(fit_mlp, fit_args)

    # Compute detector
    det_args <- c(list(y = yj, fit_mlp_res = fit_res, w = w), detector_params)
    det.lst[[j]] <- do.call(calc_detector, det_args)
  }

  # ------------------------------------------------------------
  # 2. Combine detectors into joint detector
  # ------------------------------------------------------------
  comb_args <- c(list(det.lst = det.lst, method = method), combine_params)
  comb <- do.call(combine_detectors, comb_args)

  joint.det <- comb$joint
  contrib.mat <- comb$contrib   # n × p matrix

  # ------------------------------------------------------------
  # 3. Detect changepoints on joint detector
  # ------------------------------------------------------------
  ecdf_args <- c(list(diff = joint.det, w = w), ecdf_params)
  ecdf_out <- do.call(detect_cp_ecdf, ecdf_args)

  cps <- ecdf_out$changepoints

  # ------------------------------------------------------------
  # 4. Contribution of each dimension at each CP
  # ------------------------------------------------------------
  if (length(cps) > 0) {
    cp_contrib <- contrib.mat[cps, , drop = FALSE]
    rownames(cp_contrib) <- paste0("cp_", cps)
  } else {
    cp_contrib <- matrix(0, nrow = 0, ncol = p)
  }

  # ------------------------------------------------------------
  # 5. Return full synchronized output
  # ------------------------------------------------------------
  list(
    detectors        = det.lst,
    joint_detector   = joint.det,
    contributions    = contrib.mat,
    changepoints     = cps,
    cp_contributions = cp_contrib,
    ecdf_output      = ecdf_out
  )
}

#' Simulate a Piecewise-Constant Signal with Smooth Trend and Noise
#'
#' Generate a univariate signal composed of:
#' - a smooth deterministic component,
#' - a piecewise-constant step component defined by changepoint indices,
#' - additive Gaussian noise.
#'
#' This function is useful for benchmarking changepoint detection algorithms
#' and reproducing controlled simulation studies.
#'
#' @param n Integer. Length of the signal.
#' @param domain Numeric vector of length 2 giving the range of the time axis.
#' @param changepoints_idx Integer vector of changepoint locations (indices).
#'   Must lie strictly inside \eqn{(1, n)}.
#' @param shift_sizes Numeric vector giving the mean level of each segment.
#'   Must have length equal to \code{length(changepoints_idx) + 1}.
#' @param noise_sd Numeric. Standard deviation of the Gaussian noise.
#' @param smooth_fun Function. A function \eqn{f(t)} defining the smooth trend.
#' @param seed Optional integer. If supplied, sets the random seed for reproducibility.
#'
#' @return A list with components:
#' \describe{
#'   \item{t}{Time grid of length \eqn{n}.}
#'   \item{smooth}{Smooth component \eqn{f(t)}.}
#'   \item{step}{Piecewise-constant step component.}
#'   \item{z}{Final noisy signal.}
#'   \item{changepoints_idx}{Sorted changepoint indices.}
#'   \item{shift_sizes}{Segment means.}
#'   \item{params}{List of simulation parameters.}
#' }
#'
#' @examples
#' sim <- simulate_piecewise_signal_idx(
#'   n = 1000,
#'   changepoints_idx = c(300, 700),
#'   shift_sizes = c(0.5, -0.3, 0.7),
#'   noise_sd = 0.05,
#'   seed = 123
#' )
#'
#' @export
simulate_piecewise_signal_idx <- function(
    n = 1000,
    domain = c(-4, 4),
    changepoints_idx = c(300, 700),
    shift_sizes = c(0.5, -0.3, 0.7),
    noise_sd = 0.04,
    smooth_fun = function(x) 0.01 * (3*x/2 - x^3/2),
    seed = NULL
) {
  if (!is.null(seed)) set.seed(seed)

  # Create domain
  t <- seq(domain[1], domain[2], length.out = n)

  # Validate changepoints
  if (any(changepoints_idx <= 1 | changepoints_idx >= n)) {
    stop("changepoints_idx must be inside (1, n).")
  }
  if (!is.numeric(changepoints_idx) || any(changepoints_idx %% 1 != 0)) {
    stop("changepoints_idx must be integer indices.")
  }

  # Sort indices to avoid user errors
  changepoints_idx <- sort(changepoints_idx)

  # Number of segments
  n_segments <- length(changepoints_idx) + 1

  if (length(shift_sizes) != n_segments) {
    stop(sprintf(
      "Length of 'shift_sizes' (%d) must equal number of segments (%d).",
      length(shift_sizes), n_segments
    ))
  }

  # Step function container
  g <- numeric(n)

  # Build index cuts
  cuts_idx <- c(0, changepoints_idx, n)

  for (i in seq_along(shift_sizes)) {
    idx <- (cuts_idx[i] + 1):cuts_idx[i + 1]
    g[idx] <- shift_sizes[i]
  }

  # Smooth continuous part
  f <- smooth_fun(t)

  # Final signal
  z <- f + g + rnorm(n, sd = noise_sd)

  list(
    t = t,
    smooth = f,
    step = g,
    z = z,
    changepoints_idx = changepoints_idx,
    shift_sizes = shift_sizes,
    params = list(noise_sd = noise_sd, domain = domain, n.val = n)
  )
}

#' Two-Sided Moving Average Smoothing
#'
#' @description
#' Applies a simple two-sided moving average smoother to a numeric vector.
#' This is used throughout the scanCP pipeline for stabilizing detector
#' statistics and spacing curves.
#'
#' @param x Numeric vector to smooth.
#' @param n Integer. Total number of neighbors used in the moving average.
#'   Must be >= 1.
#' @param circular Logical. If TRUE, the smoothing wraps around the ends.
#'
#' @return A numeric vector of the same length as `x` containing the smoothed
#'   values. Endpoints may be `NA` if `circular = FALSE`.
#'
#' @details
#' This function is a thin wrapper around `stats::filter()` with a symmetric
#' moving average kernel. It is intentionally lightweight and dependency-free.
#'
#' @examples
#' x <- rnorm(100)
#' y <- ma(x, n = 5)
#'
#' @importFrom stats filter
#' @export
ma <- function(x, n = 5, circular = FALSE) {

  # Validate input
  if (!is.numeric(x)) stop("x must be numeric.")
  if (n < 1) stop("n must be >= 1.")

  # Create equal-weight kernel
  kernel <- rep(1 / n, n)

  # Apply two-sided moving average
  stats::filter(x, kernel, sides = 2, circular = circular)
}

#' Full Changepoint Detection Pipeline Using Rolling MLPs
#'
#' @description
#' Runs the complete changepoint detection pipeline on a univariate signal.
#' This includes rolling-window MLP fitting, detector construction,
#' thresholding, structural decomposition, and changepoint extraction.
#'
#' @param y Numeric vector. The input signal.
#' @param w Integer. Window size for the small-window MLP used in
#'   \code{\link{fit_mlp}}. The large-window MLP automatically uses
#'   window size \code{2w}.
#' @param ma_window Integer. Window size for the moving-average smoothing
#'   applied to the detector statistic. Defaults to \code{w}.
#' @param threshold Character or numeric. If \code{"auto"}, the threshold is
#'   selected automatically using spacing-curve analysis. Otherwise, a numeric
#'   threshold may be supplied directly.
#' @param threshold_tails Numeric vector of length 2. Left and right tail
#'   cutoffs used when estimating the automatic threshold. Defaults to
#'   \code{c(0.2, 0.95)}.
#' @param min_cp_distance Integer. Minimum separation (in indices) between
#'   detected changepoints. Defaults to \code{2 * w}.
#' @param margin Integer. Margin used during local refinement of changepoint
#'   locations. Defaults to \code{floor(w / 2)}.
#' @param use_abs_det Logical. Whether to use the absolute value of the
#'   detector statistic when identifying changepoints.
#' @param mlp_control A named list of hyperparameters passed directly to
#'   \code{\link{fit_mlp}}. Any subset may be supplied; unspecified values
#'   fall back to defaults defined in \code{fit_mlp()}.
#' @param parallel Logical. If \code{TRUE}, rolling MLP fitting is
#'   performed in parallel via \code{\link{fit_mlp}}. If \code{FALSE}, all
#'   computation is performed serially. Setting \code{parallel = FALSE} is
#'   recommended for CRAN checks and for systems without parallel support.
#'
#' @return A list containing:
#' \describe{
#'   \item{\code{changepoints}}{Estimated changepoint locations.}
#'   \item{\code{detector}}{The detector statistic computed by
#'         \code{\link{calc_detector}}.}
#'   \item{\code{fit_mlp}}{Rolling MLP fit results returned by
#'         \code{\link{fit_mlp}}.}
#'   \item{\code{decomposition}}{Full structural decomposition returned by
#'         \code{\link{decompose_signal_core}}.}
#' }
#'
#' @details
#' The pipeline proceeds in three main stages:
#' \enumerate{
#'   \item Rolling-window MLP fitting via \code{\link{fit_mlp}}.
#'   \item Detector construction via \code{\link{calc_detector}}.
#'   \item Structural decomposition and changepoint extraction via
#'         \code{\link{decompose_signal_core}}.
#' }
#'
#' Detected changepoints are printed for user convenience and returned as part
#' of the output list. Parallel computation is optional and controlled by the
#' \code{parallel} argument.
#'
#'
#' @examples
#' # Minimal example
#' set.seed(1)
#' y <- c(rnorm(200, 0), rnorm(200, 3))
#'
#' \donttest{
#'   # Full pipeline (parallel disabled for CRAN)
#'   res <- scan_cp(y, w = 20, parallel = FALSE)
#' }
#'
#'
#' @export
scan_cp <- function(
    y,
    w = 100,
    ma_window = w,
    threshold = "auto",
    threshold_tails = c(0.2, 0.95),
    min_cp_distance = 2 * w,
    margin = floor(w / 2),
    use_abs_det = TRUE,
    mlp_control = list(),
    parallel = FALSE
) {

  # 1. Fit rolling MLPs
  fit_args <- c(list(vec = y, w = w, parallel = parallel), mlp_control)
  fit_res <- do.call(fit_mlp, fit_args)

  # 2. Compute detector
  det <- calc_detector(
    y = y,
    fit_mlp_res = fit_res,
    w = w,
    ma_window = ma_window,
    use_abs_det = use_abs_det
  )

  # 3. Structural decomposition
  core <- decompose_signal_core(
    y = y,
    detector = det,
    w = w,
    ma_window = ma_window,
    threshold = threshold,
    right_tail_cutoff = threshold_tails[2],
    left_tail_cutoff  = threshold_tails[1],
    min_cp_distance = min_cp_distance,
    margin = margin,
    use_abs_det = use_abs_det
  )

  # 4. Extract CPs at top level
  cps <- core$changepoints

  # 5. Print CPs for user convenience
  if (length(cps) == 0) {
    message("No changepoints detected.")
  } else {
    message("Detected changepoints at: ", paste(cps, collapse = ", "))
  }

  # 6. Return structured output
  list(
    changepoints   = cps,
    detector       = det,
    fit_mlp        = fit_res,
    decomposition  = core
  )
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_cp_multi_sync.R
\name{detect_cp_ecdf}
\alias{detect_cp_ecdf}
\title{ECDF-Based Changepoint Detection for a Detector Statistic}
\usage{
detect_cp_ecdf(
  diff,
  w = 200,
  ma_window = 100,
  right_tail_cutoff = 0.95,
  left_tail_cutoff = 0.6,
  threshold = "auto",
  circular = FALSE
)
}
\arguments{
\item{diff}{Numeric detector vector.}

\item{w}{Integer. Window size used in detector construction.}

\item{ma_window}{Integer. Moving-average window for smoothing.}

\item{right_tail_cutoff}{Numeric. ECDF upper cutoff.}

\item{left_tail_cutoff}{Numeric. ECDF lower cutoff.}

\item{threshold}{"auto" or numeric ECDF threshold.}

\item{circular}{Logical. Whether smoothing wraps around.}
}
\value{
A list containing:
\describe{
\item{changepoints}{Raw changepoint indices.}
\item{ecdf_values}{ECDF values of selected maxima.}
\item{threshold_used}{Final threshold.}
\item{local_maxima}{Matrix of local maxima.}
\item{smoothed_detector}{Smoothed detector.}
\item{ecdf_spacing}{Spacing curve.}
}
}
\description{
Applies the ECDF + spacing-curve thresholding method to a detector vector
and returns raw changepoints (no correction).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smoothing.R
\name{ma}
\alias{ma}
\title{Two-Sided Moving Average Smoothing}
\usage{
ma(x, n = 5, circular = FALSE)
}
\arguments{
\item{x}{Numeric vector to smooth.}

\item{n}{Integer. Total number of neighbors used in the moving average.
Must be >= 1.}

\item{circular}{Logical. If TRUE, the smoothing wraps around the ends.}
}
\value{
A numeric vector of the same length as \code{x} containing the smoothed
values. Endpoints may be \code{NA} if \code{circular = FALSE}.
}
\description{
Applies a simple two-sided moving average smoother to a numeric vector.
This is used throughout the scanCP pipeline for stabilizing detector
statistics and spacing curves.
}
\details{
This function is a thin wrapper around \code{stats::filter()} with a symmetric
moving average kernel. It is intentionally lightweight and dependency-free.
}
\examples{
x <- rnorm(100)
y <- ma(x, n = 5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decomposition_core.R
\name{fit_global_mlp}
\alias{fit_global_mlp}
\title{Fit Global MLP and Compute Residual Diagnostics}
\usage{
fit_global_mlp(corrected_signal)
}
\arguments{
\item{corrected_signal}{Numeric vector. Output of \code{\link[=decompose_signal_core]{decompose_signal_core()}}.}
}
\value{
A list containing:
\describe{
\item{smooth_curve}{Global MLP smooth fit.}
\item{residual}{Residuals from the smooth fit.}
\item{mlp_model}{The fitted RSNNS MLP model.}
}
}
\description{
Fits a global MLP model to a corrected signal and computes:
\itemize{
\item smooth trend estimate
\item residuals
}

This function does NOT compute a piecewise-constant component.
That is handled entirely in \code{decompose_signal_core()}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mlp_fit.R
\name{fit_mlp}
\alias{fit_mlp}
\title{Fit Rolling MLP Models for Change Point Detection}
\usage{
fit_mlp(vec, w = 100, mlp_control = list(), parallel = FALSE)
}
\arguments{
\item{vec}{Numeric vector. The input signal.}

\item{w}{Integer. Window size for the small-window MLP. The large-window
MLP automatically uses window size \code{2w}.}

\item{mlp_control}{A named list of MLP hyperparameters. Any subset may be
supplied; unspecified values fall back to defaults. Supported fields:
\describe{
\item{\code{act1}}{Activation function for the small-window MLP
(\code{"Act_Logistic"}).}
\item{\code{act2}}{Activation function for the large-window MLP
(\code{"Act_Logistic"}).}
\item{\code{lr1}}{Learning rate for the small-window MLP
(default: \code{0.001}).}
\item{\code{lr2}}{Learning learning rate for the large-window MLP
(default: \code{0.001}).}
\item{\code{hl1}}{Number of hidden units for the small-window MLP
(default: \code{6}).}
\item{\code{hl2}}{Number of hidden units for the large-window MLP
(default: \code{8}).}
\item{\code{ep1}}{Training epochs for the small-window MLP
(default: \code{300}).}
\item{\code{ep2}}{Training epochs for the large-window MLP
(default: \code{600}).}
}}

\item{parallel}{Logical. If \code{TRUE}, rolling-window MLPs are
fitted using parallel computation via \pkg{parallel}, \pkg{foreach},
and \pkg{doSNOW}. If \code{FALSE}, all computation is performed
serially. Setting \code{parallel = FALSE} is recommended for CRAN checks
and for systems where parallel backends are unavailable.}
}
\value{
A list with two elements:
\describe{
\item{\code{small}}{List of fitted values for each rolling window of size \code{w}.}
\item{\code{large}}{List of fitted values for each rolling window of size \code{2w}.}
}
}
\description{
Fits two multilayer perceptron (MLP) models over rolling windows of a
univariate signal. A small-window MLP (window size \code{w}) captures
local structure, while a large-window MLP (window size \code{2w})
captures broader trends. Their residual behavior is used by
\code{\link{calc_detector}} to construct the changepoint detector.

All MLP hyperparameters are supplied through a unified
\code{mlp_control} list for consistency with \code{\link{scan_cp}}.
}
\details{
Each rolling window is standardized before fitting. The returned matrices
contain three columns:
\enumerate{
\item standardized input \code{x}
\item fitted values \code{y_hat}
\item original indices
}

When \code{parallel = TRUE}, a cluster is created using
\code{parallel::makeCluster()}.

When \code{parallel = FALSE}, the function falls back to
serial execution using \code{\%do\%}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_cp_multi_async.R
\name{scan_cp_multi_async}
\alias{scan_cp_multi_async}
\title{Asynchronous Multivariate Changepoint Detection}
\usage{
scan_cp_multi_async(
  Y,
  w = 100,
  ma_window = w,
  threshold = "auto",
  threshold_tails = c(0.6, 0.95),
  min_cp_distance = 2 * w,
  margin = floor(w/2),
  use_abs_det = TRUE,
  mlp_control = list()
)
}
\arguments{
\item{Y}{Numeric matrix (n × p). Each column is a signal dimension.}

\item{w}{Integer. Window size for rolling MLPs and detector construction.
Defaults to 100.}

\item{ma_window}{Integer. Moving-average smoothing window for the detector.
Defaults to \code{w}.}

\item{threshold}{Either \code{"auto"} (default) or a numeric ECDF threshold.}

\item{threshold_tails}{Numeric vector of length 2 giving tail cutoff values
for automatic thresholding.
Defaults to \code{c(0.6, 0.95)}.}

\item{min_cp_distance}{Integer. Minimum distance between detected
changepoints. Defaults to \code{2 * w}.}

\item{margin}{Integer. Local refinement margin. Defaults to \code{floor(w/2)}.}

\item{use_abs_det}{Logical. Whether to use \code{abs(detector)} before
smoothing. Defaults to \code{TRUE}.}

\item{mlp_control}{List of parameters passed to \code{\link{fit_mlp}}.}
}
\value{
A named list of length \code{p}, where each element is the output
of \code{\link{scan_cp}} applied to the corresponding column of \code{Y}.
}
\description{
Applies the univariate \code{\link{scan_cp}} pipeline independently to
each column of a multivariate signal. Each dimension is processed
separately (asynchronously), producing a list of univariate results.
}
\examples{
\donttest{
  Y <- cbind(
    x1 = c(rnorm(200), rnorm(200, 3)),
    x2 = c(rnorm(200), rnorm(200, -2))
  )

  res <- scan_cp_multi_async(Y, w = 100)
  res$x1$changepoints
  res$x2$changepoints
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitting.R
\name{best_split_free}
\alias{best_split_free}
\title{Best Split for Binary Labels (Free Assignment)}
\usage{
best_split_free(labels)
}
\arguments{
\item{labels}{Integer vector containing only 1 and 2.}
}
\value{
A list with:
\describe{
\item{index}{The optimal split location (between index and index + 1).}
\item{total_correct}{Maximum number of correctly assigned labels.}
\item{left_majority}{Majority class on the left segment.}
\item{right_majority}{Majority class on the right segment.}
\item{accuracy_if_assigning_majorities}{Proportion correctly assigned.}
}
}
\description{
Given a sequence of labels in \{1, 2\}, this function finds the best split
point that maximizes the number of correctly assigned labels on each side.
The left and right segments are each assigned their majority class.
}
\details{
For each possible split point \(i\), the function computes:
\itemize{
\item left correct = max(#1_left, #2_left)
\item right correct = max(#1_right, #2_right)
}
and selects the split maximizing their sum.
}
\examples{
labels <- c(1,1,1,2,2,2)
best_split_free(labels)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotting.R
\name{plot_mlp_fits_interactive}
\alias{plot_mlp_fits_interactive}
\title{Interactive Plot of Rolling MLP Fits}
\usage{
plot_mlp_fits_interactive(
  y,
  fit_mlp_res,
  t.chp.ind = NA,
  step = 50,
  start = 75,
  w = 100
)
}
\arguments{
\item{y}{Numeric vector. The original signal.}

\item{fit_mlp_res}{A list containing rolling MLP fit results. Expected to be
a list of two lists, each containing matrices with fitted values.}

\item{t.chp.ind}{Optional numeric vector of changepoint indices to draw as
vertical dashed lines.}

\item{step}{Integer. Step size for the slider increments.}

\item{start}{Integer. Starting index for the first model window.}

\item{w}{Integer. Window size used in the rolling MLP fitting.}
}
\value{
A Plotly object.
}
\description{
Creates an interactive Plotly visualization showing multiple rolling-window
MLP fits over a univariate signal. A slider allows the user to move through
different window positions and compare model fits dynamically.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_CP.R
\name{scan_cp}
\alias{scan_cp}
\title{Full Changepoint Detection Pipeline Using Rolling MLPs}
\usage{
scan_cp(
  y,
  w = 100,
  ma_window = w,
  threshold = "auto",
  threshold_tails = c(0.2, 0.95),
  min_cp_distance = 2 * w,
  margin = floor(w/2),
  use_abs_det = TRUE,
  mlp_control = list(),
  parallel = FALSE
)
}
\arguments{
\item{y}{Numeric vector. The input signal.}

\item{w}{Integer. Window size for the small-window MLP used in
\code{\link{fit_mlp}}. The large-window MLP automatically uses
window size \code{2w}.}

\item{ma_window}{Integer. Window size for the moving-average smoothing
applied to the detector statistic. Defaults to \code{w}.}

\item{threshold}{Character or numeric. If \code{"auto"}, the threshold is
selected automatically using spacing-curve analysis. Otherwise, a numeric
threshold may be supplied directly.}

\item{threshold_tails}{Numeric vector of length 2. Left and right tail
cutoffs used when estimating the automatic threshold. Defaults to
\code{c(0.2, 0.95)}.}

\item{min_cp_distance}{Integer. Minimum separation (in indices) between
detected changepoints. Defaults to \code{2 * w}.}

\item{margin}{Integer. Margin used during local refinement of changepoint
locations. Defaults to \code{floor(w / 2)}.}

\item{use_abs_det}{Logical. Whether to use the absolute value of the
detector statistic when identifying changepoints.}

\item{mlp_control}{A named list of hyperparameters passed directly to
\code{\link{fit_mlp}}. Any subset may be supplied; unspecified values
fall back to defaults defined in \code{fit_mlp()}.}

\item{parallel}{Logical. If \code{TRUE}, rolling MLP fitting is
performed in parallel via \code{\link{fit_mlp}}. If \code{FALSE}, all
computation is performed serially. Setting \code{parallel = FALSE} is
recommended for CRAN checks and for systems without parallel support.}
}
\value{
A list containing:
\describe{
\item{\code{changepoints}}{Estimated changepoint locations.}
\item{\code{detector}}{The detector statistic computed by
\code{\link{calc_detector}}.}
\item{\code{fit_mlp}}{Rolling MLP fit results returned by
\code{\link{fit_mlp}}.}
\item{\code{decomposition}}{Full structural decomposition returned by
\code{\link{decompose_signal_core}}.}
}
}
\description{
Runs the complete changepoint detection pipeline on a univariate signal.
This includes rolling-window MLP fitting, detector construction,
thresholding, structural decomposition, and changepoint extraction.
}
\details{
The pipeline proceeds in three main stages:
\enumerate{
\item Rolling-window MLP fitting via \code{\link{fit_mlp}}.
\item Detector construction via \code{\link{calc_detector}}.
\item Structural decomposition and changepoint extraction via
\code{\link{decompose_signal_core}}.
}

Detected changepoints are printed for user convenience and returned as part
of the output list. Parallel computation is optional and controlled by the
\code{parallel} argument.
}
\examples{
# Minimal example
set.seed(1)
y <- c(rnorm(200, 0), rnorm(200, 3))

\donttest{
  # Full pipeline (parallel disabled for CRAN)
  res <- scan_cp(y, w = 20, parallel = FALSE)
}


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decomposition_core.R
\name{decompose_signal_core}
\alias{decompose_signal_core}
\title{Core Signal Decomposition (Changepoints + Piecewise Correction)}
\usage{
decompose_signal_core(
  y,
  detector,
  w = 200,
  ma_window = 100,
  right_tail_cutoff = 0.95,
  left_tail_cutoff = 0.6,
  threshold = "auto",
  use_abs_det = TRUE,
  min_cp_distance = NULL,
  circular = FALSE,
  margin = NULL
)
}
\arguments{
\item{y}{Numeric vector. Original signal.}

\item{detector}{Numeric vector. Detector statistic produced by
\code{\link{calc_detector}}.}

\item{w}{Integer. Window size used in detector construction.}

\item{ma_window}{Integer. Moving-average smoothing window for the detector
and spacing curve. Defaults to \code{w}.}

\item{right_tail_cutoff}{Numeric. Upper ECDF cutoff for automatic
threshold selection. Defaults to \code{0.95}.}

\item{left_tail_cutoff}{Numeric. Lower ECDF cutoff for automatic
threshold selection. Defaults to \code{0.6}.}

\item{threshold}{Either \code{"auto"} (default) or a numeric ECDF
threshold for selecting significant peaks.}

\item{use_abs_det}{Logical. Whether to use \code{abs(detector)} before
peak detection. Defaults to \code{TRUE}.}

\item{min_cp_distance}{Integer. Minimum distance between detected peaks.
Defaults to \code{2 * w}.}

\item{circular}{Logical. Whether moving-average smoothing wraps around.}

\item{margin}{Integer. Refinement margin around each changepoint.
Default \code{floor(w / 2)}.}
}
\value{
A list containing:
\describe{
\item{corrected_signal}{The refined, piecewise-corrected signal.}
\item{piecewise_constant}{Estimated piecewise-constant component.}
\item{raw_correction}{Cumulative correction vector.}
\item{changepoints}{Refined changepoint locations.}
\item{changepoint_ecdf}{ECDF values of selected peaks.}
\item{shift_values}{Estimated shifts at each changepoint.}
\item{smoothed_detector}{Smoothed detector statistic.}
\item{ecdf_spacing}{Spacing-curve values (if \code{threshold = "auto"}).}
\item{threshold_used}{Final threshold applied.}
\item{local_maxima}{Matrix of detected local maxima.}
}
}
\description{
Extracts changepoints from a detector statistic using ECDF thresholding,
spacing-curve analysis, and local refinement. Produces the corrected
signal and its piecewise-constant component. This function does \emph{not}
perform global MLP smoothing.
}
\details{
The function identifies local maxima of the smoothed detector, evaluates
their significance using the ECDF of the detector, optionally computes a
spacing curve for automatic threshold selection, and refines each
changepoint using a two-cluster k-means split within a local window.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_cp_multi_sync.R
\name{combine_detectors}
\alias{combine_detectors}
\title{Combine Multiple Detector Statistics into a Joint Detector}
\usage{
combine_detectors(
  det.lst,
  method = c("L1", "L2", "max"),
  ma_window = NULL,
  circular = FALSE,
  scale_contributions = TRUE,
  scale_joint = TRUE
)
}
\arguments{
\item{det.lst}{List of numeric detector vectors (all same length).}

\item{method}{One of \code{"L1"}, \code{"L2"}, \code{"max"}.}

\item{ma_window}{Optional integer. If provided, each detector is smoothed
using \code{ma()} with this window size.}

\item{circular}{Logical. Whether smoothing wraps around.}

\item{scale_contributions}{Logical. Whether to compute contribution weights.}

\item{scale_joint}{Logical. Whether to scale the joint detector to [0,1].}
}
\value{
A list containing:
\describe{
\item{joint}{Joint detector vector.}
\item{s.mat}{Scaled detector matrix.}
\item{det.mat}{Raw detector matrix.}
\item{contrib}{Contribution weights (or NULL).}
}
}
\description{
Combines several univariate detector vectors into a single joint detector
using L1, L2, or max aggregation. Optionally applies moving-average
smoothing and computes contribution weights.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/detector.R
\name{calc_detector}
\alias{calc_detector}
\title{Compute the MLP-Based Changepoint Detector Statistic}
\usage{
calc_detector(y, fit_mlp_res, w = 100,
              ma_window = w, use_abs_det = TRUE)
}
\arguments{
\item{y}{Numeric vector. Original signal.}

\item{fit_mlp_res}{List. Output of \code{\link{fit_mlp}}, containing:
\itemize{
\item \code{small} — list of small-window MLP predictions
\item \code{large} — list of large-window MLP predictions
}}

\item{w}{Integer. Window size used for the small-window MLP.}

\item{ma_window}{Integer. Moving-average smoothing window applied to the
detector. Defaults to \code{w}.}

\item{use_abs_det}{Logical. Whether to take \code{abs(detector)} before
smoothing. Defaults to \code{TRUE}.}
}
\value{
A numeric vector of detector values of length \code{n - 2w}.
}
\description{
Computes the hybrid detector statistic used in scanCP. The detector
combines a \emph{ratio} component and a \emph{difference} component derived from
the residual sum of squares (RSS) of:
\itemize{
\item two adjacent small-window MLP fits (window size \code{w})
\item one large-window MLP fit (window size \verb{2w})
}

The detector highlights locations where the large-window fit performs
substantially worse than the two small-window fits, indicating a
potential changepoint.
}
\details{
For each index \code{i}, the detector uses:
\itemize{
\item \code{rss1}: RSS of the small-window MLP on \code{[i, i+w-1]}
\item \code{rss2}: RSS of the small-window MLP on \code{[i+w, i+2w-1]}
\item \code{rss.tot}: RSS of the large-window MLP on \code{[i, i+2w-1]}
}

The ratio component is:
\deqn{ (rss.tot + b) / (rss1 + rss2 + b) }

The difference component is:
\deqn{ rss.tot - rss1 - rss2 }

The final detector is a convex combination:
\deqn{ (1 - a) \cdot \text{ratio} + a \cdot \text{difference} }

Internal parameters \code{a}, \code{b}, and \code{scale_01} are kept
fixed at their defaults for simplicity and stability.
}
\examples{
\donttest{
  y <- rnorm(300)
  fit <- fit_mlp(y, w = 100)
  det <- calc_detector(y, fit, w = 100)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitting.R
\name{select_best_spike}
\alias{select_best_spike}
\title{Select the Most Significant Spike in an ECDF Spacing Curve}
\usage{
select_best_spike(s,
                right_tail_cutoff = 0.95,
                left_tail_cutoff  = 0.6)
}
\arguments{
\item{s}{Numeric vector. Smoothed spacing curve.}

\item{right_tail_cutoff}{Numeric in (0,1). Exclude spikes with ECDF
probability above this value (typically near 1).}

\item{left_tail_cutoff}{Numeric in (0,1). Exclude spikes with ECDF
probability below this value (to avoid trivial early spikes).}
}
\value{
A numeric value in (0,1) representing the selected ECDF threshold,
or \code{NA} if no valid spike is found.
}
\description{
Identifies the most prominent spike in a spacing curve derived from the
ECDF of a smoothed detector statistic. This is used to automatically
determine a significance threshold for changepoint selection.
}
\details{
Uses \code{pracma::findpeaks()} to identify local maxima. Prominence is computed
as:
\deqn{ peak_{height} - \max(left_{base}, right_{base}) }{}

The spike with the largest prominence within the allowed ECDF range is
selected.
}
\examples{
s <- runif(100)
select_best_spike(s)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_signal.R
\name{simulate_piecewise_signal_idx}
\alias{simulate_piecewise_signal_idx}
\title{Simulate a Piecewise-Constant Signal with Smooth Trend and Noise}
\usage{
simulate_piecewise_signal_idx(
  n = 1000,
  domain = c(-4, 4),
  changepoints_idx = c(300, 700),
  shift_sizes = c(0.5, -0.3, 0.7),
  noise_sd = 0.04,
  smooth_fun = function(x) 0.01 * (3 * x/2 - x^3/2),
  seed = NULL
)
}
\arguments{
\item{n}{Integer. Length of the signal.}

\item{domain}{Numeric vector of length 2 giving the range of the time axis.}

\item{changepoints_idx}{Integer vector of changepoint locations (indices).
Must lie strictly inside \eqn{(1, n)}.}

\item{shift_sizes}{Numeric vector giving the mean level of each segment.
Must have length equal to \code{length(changepoints_idx) + 1}.}

\item{noise_sd}{Numeric. Standard deviation of the Gaussian noise.}

\item{smooth_fun}{Function. A function \eqn{f(t)} defining the smooth trend.}

\item{seed}{Optional integer. If supplied, sets the random seed for reproducibility.}
}
\value{
A list with components:
\describe{
\item{t}{Time grid of length \eqn{n}.}
\item{smooth}{Smooth component \eqn{f(t)}.}
\item{step}{Piecewise-constant step component.}
\item{z}{Final noisy signal.}
\item{changepoints_idx}{Sorted changepoint indices.}
\item{shift_sizes}{Segment means.}
\item{params}{List of simulation parameters.}
}
}
\description{
Generate a univariate signal composed of:
\itemize{
\item a smooth deterministic component,
\item a piecewise-constant step component defined by changepoint indices,
\item additive Gaussian noise.
}
}
\details{
This function is useful for benchmarking changepoint detection algorithms
and reproducing controlled simulation studies.
}
\examples{
sim <- simulate_piecewise_signal_idx(
  n = 1000,
  changepoints_idx = c(300, 700),
  shift_sizes = c(0.5, -0.3, 0.7),
  noise_sd = 0.05,
  seed = 123
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan_cp_multi_sync.R
\name{scan_cp_multi_sync}
\alias{scan_cp_multi_sync}
\title{Multivariate Synchronized Changepoint Detection}
\usage{
scan_cp_multi_sync(
  Y,
  w = 100,
  method = c("L1", "L2", "max"),
  mlp_params = list(),
  detector_params = list(),
  combine_params = list(),
  ecdf_params = list()
)
}
\arguments{
\item{Y}{Numeric matrix (n × p). Each column is a signal dimension.}

\item{w}{Integer. Window size for rolling MLPs and detector.}

\item{method}{Character. Combination method: "L1", "L2", or "max".}

\item{mlp_params}{List of parameters passed to \code{fit_mlp()}.}

\item{detector_params}{List of parameters passed to \code{calc_detector()}.}

\item{combine_params}{List of parameters passed to \code{combine_detectors()}.}

\item{ecdf_params}{List of parameters passed to \code{detect_cp_ecdf()}.}
}
\value{
A list containing:
\describe{
\item{detectors}{List of univariate detectors (one per dimension).}
\item{joint_detector}{Joint detector vector.}
\item{contributions}{Matrix of per-dimension contributions.}
\item{changepoints}{Detected synchronized changepoints.}
\item{cp_contributions}{Contribution of each dimension at each CP.}
\item{ecdf_output}{Full output of \code{detect_cp_ecdf()}.}
}
}
\description{
Implements the synchronized multivariate changepoint pipeline:
\enumerate{
\item Fit univariate MLP-based detectors for each dimension.
\item Combine detectors into a joint detector using L1/L2/max.
\item Detect changepoints on the joint detector.
\item Compute per-dimension contributions at each detected CP.
}
}
